# KCParser — created by Kyle James Catterall

import math
import re

RE_REAL_NUM = r"[+-]?([0-9]*[.])?[0-9]+"

RECOGNISED_FUNCTIONS = ("abs", "sqrt", "root", "ncr", "npr", "sigma", "pi", 
                        "ln", "log", "sin", "asin", "csc", "sinh", "cos", 
                        "acos", "sec", "cosh", "tan", "atan", "cot", 
                        "tanh")

CONSTANTS = {
    "PI": math.pi,
    "E": math.e,
    "EM": 0.577215664901533,
    "GOLDEN": 1.6180339887498948,
    "MM": 0.261497212847643,
    "BERNSTEIN": 0.280169499023869,
    "GKW": 0.303663002898733,
    "HSM": 0.3532363718549960,
    "OMEGA": 0.5671432904097839,
    "GD": 0.624329988543551,
    "CAHEN": 0.6434105462,
    "TP": 0.660161815846870,
    "LAPLACE": 0.662743419349182,
    "ET": 0.70258,
    "LR": 0.764223653589221,
    "BRUNPQ": 0.87058838,
    "BRUNPT": 1.902160583104,
    "CATALAN": 0.9159655941772190,
    "VISWANATH": 1.13198824,
    "APERY": 1.202056903159594,
    "CONWAY": 1.303577269034297,
    "MILLS": 1.306377883863081,
    "PLASTIC": 1.3247179572447460,
    "RS": 1.4513692348833811,
    "BACKHOUSE": 1.456074948582690,
    "PORTER": 1.467078079433975,
    "LIEB": 1.4670780794,
    "EB": 1.606695152415292,
    "NIVEN": 1.705211140105368,
    "UPC": 2.2955871493926381,
    "FEIGENBAUMA": 2.5029078750958929,
    "FEIGENBAUMD": 4.6692016091029907,
    "SIERPINSKI": 2.584981759579253,
    "KHINCHIN": 2.685452001065307,
    "FR": 2.807770242028519,
    "LEVY": 3.275822918721811,
    "RF": 3.359885666243178
}

ans = 0


class KCParser:
    def __init__(self, expr) -> None:
        
        def gpm(expression) -> str:
            # Use Regular Expressions to apply modifications to expression
            expression = expression.replace("#", "(#)")
            for match in re.finditer(
r"[0-9](\(|\[)|(\)|\])[a-z]|(\)|\])(\(|\[)|[0-9][a-z]|!([0-9]|[a-z]|(\(|\[))", 
expression):
                expr = match.group(0)
                exprm = f"{expr[0]}*{expr[1]}"
                expression = expression.replace(expr, exprm)
            return expression
        
        def constm(expression) -> str:
            # Replace constants [c] with value of key c in CONSTANTS dictionary
            if re.search(r"\[.+\]", expression):
                for match in re.finditer(r"\[(.+)\]", expression):
                    const_name = match.group(1)
                    if const_name.upper(
                        ) not in CONSTANTS and const_name != "ans":
                        raise SyntaxError(
                        f"{const_name} is not a valid constant")
                    else:
                        if const_name == "ans":
                            expression = expression.replace(
                                "[ans]", f"{ans}")
                        else:
                            expression = expression.replace(
                        f"[{const_name}]", f"{CONSTANTS[const_name.upper()]}")
            return expression
        
        def cfactorial(expression) -> None:
            # Check if factorials follow factorial syntax
            for match in re.finditer(f".!", expression):
                factorial = match.group(0)
                if not re.search(r"[0-9]|\)|\]", factorial):
                    raise SyntaxError("invalid factorial")
        
        def cfunction(expression) -> None:
            # Check if functions follow function syntax
            for match in re.finditer(f"[a-z]+", expression):
                func = match.group(0)
                if func not in RECOGNISED_FUNCTIONS:
                    raise SyntaxError(
                        f"{func} is not a valid function")
                if not expression[match.span()[1]] == "(":
                    raise SyntaxError(f"missing ( after {func}")
                if expression.count("(") != expression.count(")"):
                    raise SyntaxError(f"unequal number of ( and )")
                if "sigma" in expression and not (
                    temp := re.search(
                        r"sigma\([+-]?[0-9]+,[+-]?[0-9]+,.+\)", 
                        expression)):
                    raise SyntaxError("invalid sigma syntax")
                if "pi" in expression and not (
                    temp := re.search(
                        r"pi\([+-]?[0-9]+,[+-]?[0-9]+,.+\)", 
                        expression)):
                    raise SyntaxError("invalid pi syntax")
        
        self.expr = f"({expr})".strip().replace(" ", "")
        self.expr = gpm(self.expr)
        self.expr = constm(self.expr)
        self.expr = self.expr.replace("(#)", "#")
        cfactorial(self.expr)
        cfunction(self.expr)
    
    def parse(self) -> str:
        # Parse function sequence
        self.__eval_brackets()
        self.expr = self.__calc(self.expr)
        return self.expr
    
    def __eval_brackets(self) -> None:
        # Evaluate expressions within brackets, most nested first
        while "(" in self.expr:
            match = re.search(r"\(((?:[^()]+|R)+)\)", self.expr)
            inner_most = match.group(1)
            inner_most_p = inner_most
            for char in ["+", "*", "."]:
                if char in inner_most_p:
                    inner_most_p = inner_most_p.replace(char, fr"\{char}")
            if func := re.search(fr"[a-z]+\({inner_most_p}\)", self.expr):
                func = func.group(0)
                func_val = str(self.__calc(func))
                self.expr = self.expr.replace(func, func_val)            
            else:
                if "#" in inner_most:
                    result = inner_most
                else:
                    result = str(self.__calc(inner_most))
                self.expr = self.expr.replace(f"({inner_most})", f"{result}")
            if self.expr.count("(") == len(re.findall(r"[a-z]\(", self.expr)):
                break
    
    def __calc(self, expression) -> str:
        # Calculate expressions.
        global ans
        
        # Factorial (additional operator)
        if match := re.search(fr"({RE_REAL_NUM})!", expression):
            factorial = match.group(0)
            n = float(match.group(1))
            result = math.gamma(n + 1)
            expression = expression.replace(factorial, f"{result}")
        
        # Absolute
        if match := re.search(fr"abs\(({RE_REAL_NUM})\)", expression):
            absv = match.group(0)
            n = float(match.group(1))
            result = abs(n)
            expression = expression.replace(absv, f"{result}")
        
        # Sqrt
        if match := re.search(fr"sqrt\(({RE_REAL_NUM})\)", expression):
            sqrt = match.group(0)
            n = float(match.group(1))
            result = n ** 0.5
            expression = expression.replace(sqrt, f"{result}")
        
        # xth-Root
        if match := re.search(
            fr"root\(({RE_REAL_NUM}),({RE_REAL_NUM})\)", expression):
            xroot = match.group(0)
            n = float(match.group(1))
            x = float(match.group(3))
            result = n ** (1 / x)
            expression = expression.replace(xroot, f"{result}")
        
        # n C r
        if match := re.search(r"ncr\(([0-9]+),([0-9]+)\)", expression):
            ncr = match.group(0)
            n = int(match.group(1))
            r = int(match.group(2))
            if r > n:
                raise ValueError("for ncr, n >= r")
            result, k = 1, 1
            if r != 0:
                while r:
                    result *= n
                    k *= r
                    m = math.gcd(result, k)
                    result //= m    
                    k //= m
                    n -= 1
                    r -= 1
            else:
                result = 1
            expression = expression.replace(ncr, f"{result}")
        
        # n P r
        if match := re.search(r"npr\(([0-9]+),([0-9]+)\)", expression):
            npr = match.group(0)
            n = int(match.group(1))
            r = int(match.group(2))
            if r > n:
                raise ValueError("for npr, n >= r")
            result = math.floor(self.__calc(f"{n}!") / self.__calc(f"{n - r}!"))
            expression = expression.replace(npr, f"{result}")
        
        # Sigma
        if match := re.search(
            r"sigma\(([+-]?[0-9]+),([+-]?[0-9]+),(.+)\)", expression):
            sigma = match.group(0)
            start = match.group(1)
            end = match.group(2)
            expr = match.group(3)
            if int(start) > int(end):
                start, end = end, start
            result = 0
            for x in range(int(start), int(end) + 1):
                cexpr = expr.replace("#", str(x))
                result += self.__calc(cexpr)
            expression = expression.replace(sigma, f"{result}")
        
        # Pi
        if match := re.search(
            r"pi\(([+-]?[0-9]+),([+-]?[0-9]+),(.+)\)", expression):
            pi = match.group(0)
            start = match.group(1)
            end = match.group(2)
            expr = match.group(3)
            if int(start) > int(end):
                start, end = end, start
            result = 1
            for x in range(int(start), int(end) + 1):
                cexpr = expr.replace("#", str(x))
                result *= self.__calc(cexpr)
            expression = expression.replace(pi, f"{result}")
        
        # Natural Logarithm of x
        if match := re.search(fr"ln\(({RE_REAL_NUM})\)", expression):
            ln = match.group(0)
            x = float(match.group(1))
            result = math.log(x)
            expression = expression.replace(ln, f"{result}")
        
        # Logarithm of x to the base b
        if match := re.search(
            fr"log\(({RE_REAL_NUM}),({RE_REAL_NUM})\)", expression):
            log = match.group(0)
            x = float(match.group(1))
            b = float(match.group(3))
            result = math.log(x, b)
            expression = expression.replace(log, f"{result}")
        
        # Sine of x
        if match := re.search(fr"sin\(({RE_REAL_NUM})\)", expression):
            sin = match.group(0)
            x = float(match.group(1))
            result = math.sin(x)
            expression = expression.replace(sin, f"{result}")
        
        # Arcsine of x
        if match := re.search(fr"asin\(({RE_REAL_NUM})\)", expression):
            asin = match.group(0)
            x = float(match.group(1))
            result = math.asin(x)
            expression = expression.replace(asin, f"{result}")
        
        # Cosecant of x
        if match := re.search(fr"csc\(({RE_REAL_NUM})\)", expression):
            csc = match.group(0)
            x = float(match.group(1))
            result = 1 / math.sin(x)
            expression = expression.replace(csc, f"{result}")
        
        # Hyperbolic sine of x
        if match := re.search(fr"sinh\(({RE_REAL_NUM})\)", expression):
            sinh = match.group(0)
            x = float(match.group(1))
            result = math.sinh(x)
            expression = expression.replace(sinh, f"{result}")
        
        # Cosine of x
        if match := re.search(fr"cos\(({RE_REAL_NUM})\)", expression):
            cos = match.group(0)
            x = float(match.group(1))
            result = math.cos(x)
            expression = expression.replace(cos, f"{result}")
        
        # Arccosine of x
        if match := re.search(fr"acos\(({RE_REAL_NUM})\)", expression):
            acos = match.group(0)
            x = float(match.group(1))
            result = math.acos(x)
            expression = expression.replace(acos, f"{result}")
        
        # Secant of x
        if match := re.search(fr"sec\(({RE_REAL_NUM})\)", expression):
            sec = match.group(0)
            x = float(match.group(1))
            result = 1 / math.cos(x)
            expression = expression.replace(sec, f"{result}")
        
        # Hyperbolic cosine of x
        if match := re.search(fr"cosh\(({RE_REAL_NUM})\)", expression):
            cosh = match.group(0)
            x = float(match.group(1))
            result = math.cosh(x)
            expression = expression.replace(cosh, f"{result}")
        
        # Tangent of x
        if match := re.search(fr"tan\(({RE_REAL_NUM})\)", expression):
            tan = match.group(0)
            x = float(match.group(1))
            result = math.tan(x)
            expression = expression.replace(tan, f"{result}")
        
        # Arctan of x
        if match := re.search(fr"atan\(({RE_REAL_NUM})\)", expression):
            atan = match.group(0)
            x = float(match.group(1))
            result = math.atan(x)
            expression = expression.replace(atan, f"{result}")
        
        # Cotangent of x
        if match := re.search(fr"cot\(({RE_REAL_NUM})\)", expression):
            cot = match.group(0)
            x = float(match.group(1))
            result = 1 / math.tan(x)
            expression = expression.replace(cot, f"{result}")
        
        # Hyperbolic tangent of x
        if match := re.search(fr"tanh\(({RE_REAL_NUM})\)", expression):
            tanh = match.group(0)
            x = float(match.group(1))
            result = math.tanh(x)
            expression = expression.replace(tanh, f"{result}")
        
        
        ans = eval(expression)
        return ans
    
    # Overload operators
    
    def __add__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}+{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}+{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __sub__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}-{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}-{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __mul__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}*{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}*{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __pow__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}**{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}**{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __truediv__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}/{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}/{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __floordiv__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}//{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}//{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")

    def __mod__(self, expression):
        if isinstance(expression, str):
            return KCParser(f"{self.expr}%{expression}")
        if isinstance(expression, KCParser):
            return KCParser(f"{self.expr}%{expression.expr}")
        raise TypeError(f"{expression} is not of class str or KCParser")
            

